@@ -22,13 +22,14 @@
 
 #include <btrfsutil.h>
 
-#include "ctree.h"
+#include "kernel-shared/ctree.h"
 #include "ioctl.h"
-
+#include "common/open-utils.h"
 #include "cmds/commands.h"
 #include "qgroup.h"
 #include "common/utils.h"
 #include "common/help.h"
+#include "common/units.h"
 
 static const char * const qgroup_cmd_group_usage[] = {
 	"btrfs qgroup <command> [options] <path>",
@@ -45,34 +46,30 @@ static int _cmd_qgroup_assign(const struct cmd_struct *cmd, int assign,
 	struct btrfs_ioctl_qgroup_assign_args args;
 	DIR *dirstream = NULL;
 
-	if (assign) {
-		optind = 0;
-		while (1) {
-			enum { GETOPT_VAL_RESCAN = 256, GETOPT_VAL_NO_RESCAN };
-			static const struct option long_options[] = {
-				{ "rescan", no_argument, NULL,
-					GETOPT_VAL_RESCAN },
-				{ "no-rescan", no_argument, NULL,
-					GETOPT_VAL_NO_RESCAN },
-				{ NULL, 0, NULL, 0 }
-			};
-			int c = getopt_long(argc, argv, "", long_options, NULL);
+	optind = 0;
+	while (1) {
+		enum { GETOPT_VAL_RESCAN = 256, GETOPT_VAL_NO_RESCAN };
+		static const struct option long_options[] = {
+			{ "rescan", no_argument, NULL, GETOPT_VAL_RESCAN },
+			{ "no-rescan", no_argument, NULL, GETOPT_VAL_NO_RESCAN },
+			{ NULL, 0, NULL, 0 }
+		};
+		int c;
 
-			if (c < 0)
-				break;
-			switch (c) {
-			case GETOPT_VAL_RESCAN:
-				rescan = true;
-				break;
-			case GETOPT_VAL_NO_RESCAN:
-				rescan = false;
-				break;
-			default:
-				usage_unknown_option(cmd, argv);
-			}
+		c = getopt_long(argc, argv, "", long_options, NULL);
+		if (c < 0)
+			break;
+
+		switch (c) {
+		case GETOPT_VAL_RESCAN:
+			rescan = true;
+			break;
+		case GETOPT_VAL_NO_RESCAN:
+			rescan = false;
+			break;
+		default:
+			usage_unknown_option(cmd, argv);
 		}
-	} else {
-		clean_args_no_options(cmd, argc, argv);
 	}
 
 	if (check_argc_exact(argc - optind, 3))
@@ -98,7 +95,9 @@ static int _cmd_qgroup_assign(const struct cmd_struct *cmd, int assign,
 
 	ret = ioctl(fd, BTRFS_IOC_QGROUP_ASSIGN, &args);
 	if (ret < 0) {
-		error("unable to assign quota group: %m");
+		error("unable to assign quota group: %s",
+				errno == ENOTCONN ? "quota not enabled"
+						: strerror(errno));
 		close_file_or_dir(fd, dirstream);
 		return 1;
 	}
@@ -152,8 +151,10 @@ static int _cmd_qgroup_create(int create, int argc, char **argv)
 	ret = ioctl(fd, BTRFS_IOC_QGROUP_CREATE, &args);
 	close_file_or_dir(fd, dirstream);
 	if (ret < 0) {
-		error("unable to %s quota group: %m",
-			create ? "create":"destroy");
+		error("unable to %s quota group: %s",
+			create ? "create":"destroy",
+				errno == ENOTCONN ? "quota not enabled"
+						: strerror(errno));
 		return 1;
 	}
 	return 0;
@@ -176,8 +177,11 @@ static int cmd_qgroup_assign(const struct cmd_struct *cmd,
 static DEFINE_SIMPLE_COMMAND(qgroup_assign, "assign");
 
 static const char * const cmd_qgroup_remove_usage[] = {
-	"btrfs qgroup remove <src> <dst> <path>",
+	"btrfs qgroup remove [options] <src> <dst> <path>",
 	"Remove a child qgroup SRC from DST.",
+	"",
+	"--rescan       schedule qutoa rescan if needed",
+	"--no-rescan    don't schedule quota rescan",
 	NULL
 };
 
@@ -408,7 +412,7 @@ static int cmd_qgroup_limit(const struct cmd_struct *cmd, int argc, char **argv)
 	if (!strcasecmp(argv[optind], "none"))
 		size = -1ULL;
 	else
-		size = parse_size(argv[optind]);
+		size = parse_size_from_string(argv[optind]);
 
 	memset(&args, 0, sizeof(args));
 	if (compressed)
@@ -447,7 +451,10 @@ static int cmd_qgroup_limit(const struct cmd_struct *cmd, int argc, char **argv)
 	ret = ioctl(fd, BTRFS_IOC_QGROUP_LIMIT, &args);
 	close_file_or_dir(fd, dirstream);
 	if (ret < 0) {
-		error("unable to limit requested quota group: %m");
+		error("unable to limit requested quota group: %s",
+				errno == ENOTCONN ? "quota not enabled"
+						: strerror(errno));
+
 		return 1;
 	}
 	return 0;
