@@ -18,13 +18,13 @@
 #define __BTRFS_CONVERT_SOURCE_FS_H__
 
 #include "kerncompat.h"
-#include <linux/kdev_t.h>
 #include <pthread.h>
+#include <sys/types.h>
 
 #define CONV_IMAGE_SUBVOL_OBJECTID BTRFS_FIRST_FREE_OBJECTID
 
 /*
- * Reresents a simple contiguous range.
+ * Represents a simple contiguous range.
  *
  * For multiple or non-contiguous ranges, use extent_cache_tree from
  * extent-cache.c
